@@ -20,6 +20,7 @@ import logging
 import socket
 
 from concurrent import futures
+from oslo_utils import encodeutils
 from pymemcache import client as pymemcache_client
 import six
 
@@ -44,14 +45,14 @@ def _translate_failures(func):
             return func(*args, **kwargs)
         except pymemcache_client.MemcacheUnexpectedCloseError as e:
             coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
         except (socket.timeout, socket.error,
                 socket.gaierror, socket.herror) as e:
             # TODO(harlowja): get upstream pymemcache to produce a better
             # exception for these, using socket (vs. a memcache specific
             # error) seems sorta not right and/or the best approach...
-            msg = utils.exception_message(e)
+            msg = encodeutils.exception_to_unicode(e)
             if e.errno is not None:
                 msg += " (with errno %s [%s])" % (errno.errorcode[e.errno],
                                                   e.errno)
@@ -59,7 +60,7 @@ def _translate_failures(func):
                                           msg, cause=e)
         except pymemcache_client.MemcacheError as e:
             coordination.raise_with_cause(coordination.ToozError,
-                                          utils.exception_message(e),
+                                          encodeutils.exception_to_unicode(e),
                                           cause=e)
 
     return wrapper
@@ -73,6 +74,15 @@ class MemcachedLock(locking.Lock):
         self.coord = coord
         self.timeout = timeout
 
+    def is_still_owner(self):
+        if not self.acquired:
+            return False
+        else:
+            owner = self.get_owner()
+            if owner is None:
+                return False
+            return owner == self.coord._member_id
+
     def acquire(self, blocking=True):
 
         @_retry.retry(stop_max_delay=blocking)
@@ -91,28 +101,74 @@ class MemcachedLock(locking.Lock):
 
         return _acquire()
 
+    @_translate_failures
+    def break_(self):
+        return bool(self.coord.client.delete(self.name, noreply=False))
+
     @_translate_failures
     def release(self):
-        if self.coord.client.delete(self.name, noreply=False):
-            self.coord._acquired_locks.remove(self)
-            return True
-        else:
+        if not self.acquired:
             return False
+        # NOTE(harlowja): this has the potential to delete others locks
+        # especially if this key expired before the delete/release call is
+        # triggered.
+        #
+        # For example:
+        #
+        # 1. App #1 with coordinator 'A' acquires lock "b"
+        # 2. App #1 heartbeats every 10 seconds, expiry for lock let's
+        #    say is 11 seconds.
+        # 3. App #2 with coordinator also named 'A' blocks trying to get
+        #    lock "b" (let's say it retries attempts every 0.5 seconds)
+        # 4. App #1 is running behind a little bit, tries to heartbeat but
+        #    key has expired (log message is written); at this point app #1
+        #    doesn't own the lock anymore but it doesn't know that.
+        # 5. App #2 now retries and adds the key, and now it believes it
+        #    has the lock.
+        # 6. App #1 (still believing it has the lock) calls release, and
+        #    deletes app #2 lock, app #2 now doesn't own the lock anymore
+        #    but it doesn't know that and now app #(X + 1) can get it.
+        # 7. App #2 calls release (repeat #6 as many times as desired)
+        #
+        # Sadly I don't think memcache has the primitives to actually make
+        # this work, redis does because it has lua which can check a session
+        # id and then do the delete and bail out if the session id is not
+        # as expected but memcache doesn't seem to have any equivalent
+        # capability.
+        if self not in self.coord._acquired_locks:
+            return False
+        # Do a ghetto test to see what the value is... (see above note),
+        # and how this really can't be done safely with memcache due to
+        # it being done in the client side (non-atomic).
+        value = self.coord.client.get(self.name)
+        if value != self.coord._member_id:
+            return False
+        else:
+            was_deleted = self.coord.client.delete(self.name, noreply=False)
+            if was_deleted:
+                self.coord._acquired_locks.remove(self)
+            return was_deleted
 
     @_translate_failures
     def heartbeat(self):
         """Keep the lock alive."""
-        poked = self.coord.client.touch(self.name,
-                                        expire=self.timeout,
-                                        noreply=False)
-        if not poked:
-            LOG.warn("Unable to heartbeat by updating key '%s' with extended"
-                     " expiry of %s seconds", self.name, self.timeout)
+        if self.acquired:
+            poked = self.coord.client.touch(self.name,
+                                            expire=self.timeout,
+                                            noreply=False)
+            if not poked:
+                LOG.warn("Unable to heartbeat by updating key '%s' with"
+                         " extended expiry of %s seconds", self.name,
+                         self.timeout)
 
     @_translate_failures
     def get_owner(self):
         return self.coord.client.get(self.name)
 
+    @property
+    def acquired(self):
+        return self in self.coord._acquired_locks
+
 
 class MemcachedDriver(coordination._RunWatchersMixin,
                       coordination.CoordinationDriver):
@@ -123,10 +179,26 @@ class MemcachedDriver(coordination._RunWatchersMixin,
     all of the coordination driver API(s). It stores data into memcache
     using expiries and `msgpack`_ encoded values.
 
+    General recommendations/usage considerations:
+
+    - Memcache (without different backend technology) is a **cache** enough
+      said.
+
     .. _memcached: http://memcached.org/
     .. _msgpack: http://msgpack.org/
     """
 
+    CHARACTERISTICS = (
+        coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_PROCESSES,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS,
+        coordination.Characteristics.CAUSAL,
+    )
+    """
+    Tuple of :py:class:`~tooz.coordination.Characteristics` introspectable
+    enum member(s) that can be used to interogate how this driver works.
+    """
+
     #: Key prefix attached to groups (used in name-spacing keys)
     GROUP_PREFIX = b'_TOOZ_GROUP_'
 
@@ -352,8 +424,8 @@ class MemcachedDriver(coordination._RunWatchersMixin,
         actual_group_members = {}
         for m, v in six.iteritems(group_members):
             # Never kick self from the group, we know we're alive
-            if (m == self._member_id
-               or self.client.get(self._encode_member_id(m))):
+            if (m == self._member_id or
+               self.client.get(self._encode_member_id(m))):
                 actual_group_members[m] = v
         if group_members != actual_group_members:
             # There are some dead members, update the group
@@ -458,7 +530,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
                              self.leader_timeout)
 
     @_translate_failures
-    def _run_leadership(self):
+    def run_elect_coordinator(self):
         for group_id, hooks in six.iteritems(self._hooks_elected_leader):
             # Try to grab the lock, if that fails, that means someone has it
             # already.
@@ -471,7 +543,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
 
     def run_watchers(self, timeout=None):
         result = super(MemcachedDriver, self).run_watchers(timeout=timeout)
-        self._run_leadership()
+        self.run_elect_coordinator()
         return result
 
 
@@ -484,9 +556,10 @@ class MemcachedFutureResult(coordination.CoordAsyncResult):
         try:
             return self._fut.result(timeout=timeout)
         except futures.TimeoutError as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          utils.exception_message(e),
-                                          cause=e)
+            coordination.raise_with_cause(
+                coordination.OperationTimedOut,
+                encodeutils.exception_to_unicode(e),
+                cause=e)
 
     def done(self):
         return self._fut.done()
