@@ -19,8 +19,13 @@ import copy
 
 from kazoo import client
 from kazoo import exceptions
+from kazoo.handlers import eventlet as eventlet_handler
+from kazoo.handlers import threading as threading_handler
 from kazoo.protocol import paths
+from oslo_utils import encodeutils
+from oslo_utils import strutils
 import six
+from six.moves import filter as compat_filter
 
 from tooz import coordination
 from tooz import locking
@@ -31,13 +36,40 @@ class ZooKeeperLock(locking.Lock):
     def __init__(self, name, lock):
         super(ZooKeeperLock, self).__init__(name)
         self._lock = lock
+        self._client = lock.client
+
+    def is_still_owner(self):
+        if not self.acquired:
+            return False
+        try:
+            data, _znode = self._client.get(
+                paths.join(self._lock.path, self._lock.node))
+            return data == self._lock.data
+        except (self._client.handler.timeout_exception,
+                exceptions.ConnectionLoss,
+                exceptions.ConnectionDropped,
+                exceptions.NoNodeError):
+            return False
+        except exceptions.KazooException as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "operation error: %s" % (e),
+                                          cause=e)
 
     def acquire(self, blocking=True):
-        return self._lock.acquire(blocking=bool(blocking),
-                                  timeout=blocking)
+        blocking, timeout = utils.convert_blocking(blocking)
+        return self._lock.acquire(blocking=blocking,
+                                  timeout=timeout)
 
     def release(self):
-        return self._lock.release()
+        if self.acquired:
+            self._lock.release()
+            return True
+        else:
+            return False
+
+    @property
+    def acquired(self):
+        return self._lock.is_acquired
 
 
 class BaseZooKeeperDriver(coordination.CoordinationDriver):
@@ -46,25 +78,32 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
     :param timeout: connection timeout to wait when first connecting to the
                     zookeeper server
     """
-
-    _TOOZ_NAMESPACE = b"tooz"
+    #: Default namespace when none is provided.
+    TOOZ_NAMESPACE = b"tooz"
 
     def __init__(self, member_id, parsed_url, options):
         super(BaseZooKeeperDriver, self).__init__()
+        options = utils.collapse(options, exclude=['hosts'])
+        self._options = options
         self._member_id = member_id
-        self.timeout = int(options.get('timeout', ['10'])[-1])
+        self.timeout = int(options.get('timeout', '10'))
+        self._namespace = options.get('namespace', self.TOOZ_NAMESPACE)
 
     def _start(self):
         try:
             self._coord.start(timeout=self.timeout)
         except self._coord.handler.timeout_exception as e:
-            raise coordination.ToozConnectionError("operation error: %s" % (e))
-
+            e_msg = encodeutils.exception_to_unicode(e)
+            coordination.raise_with_cause(coordination.ToozConnectionError,
+                                          "Operational error: %s" % e_msg,
+                                          cause=e)
         try:
-            self._coord.ensure_path(self.paths_join("/", self._TOOZ_NAMESPACE))
+            self._coord.ensure_path(self._paths_join("/", self._namespace))
         except exceptions.KazooException as e:
-            raise coordination.ToozError("operation error: %s" % (e))
-
+            e_msg = encodeutils.exception_to_unicode(e)
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "Operational error: %s" % e_msg,
+                                          cause=e)
         self._group_members = collections.defaultdict(set)
         self._watchers = collections.deque()
         self._leader_locks = {}
@@ -80,19 +119,25 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
     def _loads(blob):
         return utils.loads(blob)
 
-    @staticmethod
-    def _create_group_handler(async_result, timeout,
+    def _create_group_handler(self, async_result, timeout,
                               timeout_exception, group_id):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NodeExistsError:
             raise coordination.GroupAlreadyExist(group_id)
-        except exceptions.NoNodeError:
-            raise coordination.ToozError("tooz namespace has not been created")
+        except exceptions.NoNodeError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "Tooz namespace '%s' has not"
+                                          " been created" % self._namespace,
+                                          cause=e)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
 
     def create_group(self, group_id):
         group_path = self._path_group(group_id)
@@ -107,13 +152,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.NotEmptyError:
             raise coordination.GroupNotEmpty(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
 
     def delete_group(self, group_id):
         group_path = self._path_group(group_id)
@@ -128,13 +177,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NodeExistsError:
             raise coordination.MemberAlreadyExist(group_id, member_id)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
 
     def join_group(self, group_id, capabilities=b""):
         member_path = self._path_member(group_id, self._member_id)
@@ -152,11 +205,33 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+
+    def heartbeat(self):
+        # Just fetch the base path (and do nothing with it); this will
+        # force any waiting heartbeat responses to be flushed, and also
+        # ensures that the connection still works as expected...
+        base_path = self._paths_join("/", self._namespace)
+        try:
+            self._coord.get(base_path)
+        except self._timeout_exception as e:
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        except exceptions.NoNodeError:
+            pass
+        except exceptions.ZookeeperError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
 
     def leave_group(self, group_id):
         member_path = self._path_member(group_id, self._member_id)
@@ -171,16 +246,20 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             members_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         else:
             return set(m.encode('ascii') for m in members_ids)
 
     def get_members(self, group_id):
-        group_path = self.paths_join("/", self._TOOZ_NAMESPACE, group_id)
+        group_path = self._paths_join("/", self._namespace, group_id)
         async_result = self._coord.get_children_async(group_path)
         return ZooAsyncResult(async_result, self._get_members_handler,
                               timeout_exception=self._timeout_exception,
@@ -192,11 +271,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
 
     def update_capabilities(self, group_id, capabilities):
         member_path = self._path_member(group_id, self._member_id)
@@ -213,11 +296,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             capabilities = async_result.get(block=True, timeout=timeout)[0]
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         else:
             return cls._loads(capabilities)
 
@@ -229,55 +316,170 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
                               timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
-    @staticmethod
-    def _get_groups_handler(async_result, timeout, timeout_exception):
+    @classmethod
+    def _get_member_info_handler(cls, async_result, timeout,
+                                 timeout_exception, group_id,
+                                 member_id):
+        try:
+            capabilities, znode_stats = async_result.get(block=True,
+                                                         timeout=timeout)
+        except timeout_exception as e:
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        except exceptions.NoNodeError:
+            raise coordination.MemberNotJoined(group_id, member_id)
+        except exceptions.ZookeeperError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        else:
+            member_info = {
+                'capabilities': cls._loads(capabilities),
+                'created_at': utils.millis_to_datetime(znode_stats.ctime),
+                'updated_at': utils.millis_to_datetime(znode_stats.mtime)
+            }
+            return member_info
+
+    def get_member_info(self, group_id, member_id):
+        member_path = self._path_member(group_id, member_id)
+        async_result = self._coord.get_async(member_path)
+        return ZooAsyncResult(async_result,
+                              self._get_member_info_handler,
+                              timeout_exception=self._timeout_exception,
+                              group_id=group_id, member_id=self._member_id)
+
+    def _get_groups_handler(self, async_result, timeout, timeout_exception):
         try:
             group_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            raise coordination.OperationTimedOut(utils.exception_message(e))
-        except exceptions.NoNodeError:
-            raise coordination.ToozError("tooz namespace has not been created")
+            coordination.raise_with_cause(coordination.OperationTimedOut,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
+        except exceptions.NoNodeError as e:
+            coordination.raise_with_cause(coordination.ToozError,
+                                          "Tooz namespace '%s' has not"
+                                          " been created" % self._namespace,
+                                          cause=e)
         except exceptions.ZookeeperError as e:
-            raise coordination.ToozError(utils.exception_message(e))
+            coordination.raise_with_cause(coordination.ToozError,
+                                          encodeutils.exception_to_unicode(e),
+                                          cause=e)
         else:
             return set(g.encode('ascii') for g in group_ids)
 
     def get_groups(self):
-        tooz_namespace = self.paths_join("/", self._TOOZ_NAMESPACE)
+        tooz_namespace = self._paths_join("/", self._namespace)
         async_result = self._coord.get_children_async(tooz_namespace)
         return ZooAsyncResult(async_result, self._get_groups_handler,
                               timeout_exception=self._timeout_exception)
 
     def _path_group(self, group_id):
-        return self.paths_join("/", self._TOOZ_NAMESPACE, group_id)
+        return self._paths_join("/", self._namespace, group_id)
 
     def _path_member(self, group_id, member_id):
-        return self.paths_join("/", self._TOOZ_NAMESPACE,
-                               group_id, member_id)
+        return self._paths_join("/", self._namespace, group_id, member_id)
 
     @staticmethod
-    def paths_join(*args):
-        lpaths = []
+    def _paths_join(arg, *more_args):
+        """Converts paths into a string (unicode)."""
+        args = [arg]
+        args.extend(more_args)
+        cleaned_args = []
         for arg in args:
             if isinstance(arg, six.binary_type):
-                lpaths.append(arg.decode('ascii'))
+                cleaned_args.append(arg.decode('ascii'))
             else:
-                lpaths.append(arg)
-        return paths.join(*lpaths)
+                cleaned_args.append(arg)
+        return paths.join(*cleaned_args)
 
 
 class KazooDriver(BaseZooKeeperDriver):
-    """The driver using the Kazoo client against real ZooKeeper servers."""
+    """This driver uses the `kazoo`_ client against real `zookeeper`_ servers.
+
+    It **is** fully functional and implements all of the coordination
+    driver API(s). It stores data into `zookeeper`_ using znodes
+    and `msgpack`_ encoded values.
+
+    To configure the client to your liking a subset of the options defined at
+    http://kazoo.readthedocs.org/en/latest/api/client.html
+    will be extracted from the coordinator url (or any provided options),
+    so that a specific coordinator can be created that will work for you.
+
+    Currently the following options will be proxied to the contained client:
+
+    ================  ===============================  ====================
+    Name              Source                           Default
+    ================  ===============================  ====================
+    hosts             url netloc + 'hosts' option key  localhost:2181
+    timeout           'timeout' options key            10.0 (kazoo default)
+    connection_retry  'connection_retry' options key   None
+    command_retry     'command_retry' options key      None
+    randomize_hosts   'randomize_hosts' options key    True
+    ================  ===============================  ====================
+
+    .. _kazoo: http://kazoo.readthedocs.org/
+    .. _zookeeper: http://zookeeper.apache.org/
+    .. _msgpack: http://msgpack.org/
+    """
+
+    HANDLERS = {
+        'eventlet': eventlet_handler.SequentialEventletHandler,
+        'threading': threading_handler.SequentialThreadingHandler,
+    }
+    """
+    Restricted immutable dict of handler 'kinds' -> handler classes that
+    this driver can accept via 'handler' option key (the expected value for
+    this option is one of the keys in this dictionary).
+    """
+
+    CHARACTERISTICS = (
+        coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_PROCESSES,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS,
+        # Writes *always* go through a single leader process, but it may
+        # take a while for those writes to propagate to followers (and =
+        # during this time clients can read older values)...
+        coordination.Characteristics.SEQUENTIAL,
+    )
+    """
+    Tuple of :py:class:`~tooz.coordination.Characteristics` introspectable
+    enum member(s) that can be used to interogate how this driver works.
+    """
 
     def __init__(self, member_id, parsed_url, options):
         super(KazooDriver, self).__init__(member_id, parsed_url, options)
-        self._coord = self._make_client(parsed_url, options)
+        self._coord = self._make_client(parsed_url, self._options)
         self._member_id = member_id
         self._timeout_exception = self._coord.handler.timeout_exception
 
-    @classmethod
-    def _make_client(cls, parsed_url, options):
-        return client.KazooClient(hosts=parsed_url.netloc)
+    def _make_client(self, parsed_url, options):
+        # Creates a kazoo client,
+        # See: https://github.com/python-zk/kazoo/blob/2.2.1/kazoo/client.py
+        # for what options a client takes...
+        maybe_hosts = [parsed_url.netloc] + list(options.get('hosts', []))
+        hosts = list(compat_filter(None, maybe_hosts))
+        if not hosts:
+            hosts = ['localhost:2181']
+        randomize_hosts = options.get('randomize_hosts', True)
+        client_kwargs = {
+            'hosts': ",".join(hosts),
+            'timeout': float(options.get('timeout', self.timeout)),
+            'connection_retry': options.get('connection_retry'),
+            'command_retry': options.get('command_retry'),
+            'randomize_hosts': strutils.bool_from_string(randomize_hosts),
+        }
+        handler_kind = options.get('handler')
+        if handler_kind:
+            try:
+                handler_cls = self.HANDLERS[handler_kind]
+            except KeyError:
+                raise ValueError("Unknown handler '%s' requested"
+                                 " valid handlers are %s"
+                                 % (handler_kind,
+                                    sorted(self.HANDLERS.keys())))
+            client_kwargs['handler'] = handler_cls()
+        return client.KazooClient(**client_kwargs)
 
     def _watch_group(self, group_id):
         get_members_req = self.get_members(group_id)
@@ -394,27 +596,31 @@ class KazooDriver(BaseZooKeeperDriver):
         return ZooAsyncResult(None, lambda *args: leader)
 
     def get_lock(self, name):
-        return ZooKeeperLock(
-            name,
-            self._coord.Lock(
-                self.paths_join(b"/", self._TOOZ_NAMESPACE, b"locks", name),
-                self._member_id.decode('ascii')))
-
-    def run_watchers(self):
-        ret = []
-        while self._watchers:
-            cb = self._watchers.popleft()
-            ret.extend(cb())
+        z_lock = self._coord.Lock(
+            self._paths_join(b"/", self._namespace, b"locks", name),
+            self._member_id.decode('ascii'))
+        return ZooKeeperLock(name, z_lock)
 
+    def run_elect_coordinator(self):
         for group_id in six.iterkeys(self._hooks_elected_leader):
-            if self._get_group_leader_lock(group_id).acquire(blocking=False):
+            leader_lock = self._get_group_leader_lock(group_id)
+            if leader_lock.is_acquired:
+                # Previously acquired/still leader, leave it be...
+                continue
+            if leader_lock.acquire(blocking=False):
                 # We are now leader for this group
                 self._hooks_elected_leader[group_id].run(
                     coordination.LeaderElected(
                         group_id,
                         self._member_id))
 
-        return ret
+    def run_watchers(self, timeout=None):
+        results = []
+        while self._watchers:
+            cb = self._watchers.popleft()
+            results.extend(cb())
+        self.run_elect_coordinator()
+        return results
 
 
 class ZooAsyncResult(coordination.CoordAsyncResult):
